import { Inject, Provide } from '@midwayjs/core';
import { LoginType, UPwdDTO } from '../../frame/Options';
import { get, isEqual, upperFirst } from 'lodash';
import { GetModel, randomStr } from '../../frame/Utils';
import { ServiceError } from '../../error/CustomerError.error';
import { ErrorCode } from '../../error/Codes';
import { RoleService } from '../system/role.service';

@Provide()
export class LoginService {
  @Inject()
  roleService: RoleService;
  /**
   * 账密登录
   * @param data 用户名和密码
   * @param type 用户类型
   * @returns 用户信息/空值
   */
  async loginByAccount(data, type: LoginType) {
    const model = GetModel(upperFirst(type));
    const user = await model
      .findOne({ account: data.account }, '+password')
      .lean();
    if (!user) throw new ServiceError(ErrorCode.USER_NOT_FOUND);
    await this.checkAccountCanLogin(user, type);
    if (!isEqual(user.password.secret, data.password))
      throw new ServiceError(ErrorCode.BAD_PASSWORD);
    return user;
  }

  /**
   * 检查用户是否可以登录(从用户本身和角色检查)
   * @param user 用户信息
   * @param type 用户类型
   */
  async checkAccountCanLogin(user, type: LoginType) {
    /**
     * 判断是否可以使用:
     * 管理员:
     *  超级管理员无视,直接往下;
     *  普通管理员需要查看
     *    角色是否能用
     *    is_use是不是'0'
     * 其他用户;
     *  角色是否能用
     *  is_use是不是'0'
     * 普通管理员和其他用户是一种判断
     */

    if (type === 'Admin') {
      if (get(user, 'is_super') === '0') return true;
    }
    const role = await this.roleService.findOne({
      code: type,
      is_use: '0',
    });
    if (!role) throw new ServiceError(ErrorCode.ROLE_IS_DISABLED);
    if (get(user, 'is_use') === '1')
      throw new ServiceError(ErrorCode.USER_IS_DISABLED);
  }

 /**
  * 修改密码
  * @param data 修改密码对象
  * @param type 规定的登录类型
  * @returns 新密码
  */
  async updatePwd(data: UPwdDTO, type: LoginType) {
    const model = GetModel(upperFirst(type));
    const user = await model.findById(data._id);
    if (!user) new ServiceError(ErrorCode.USER_NOT_FOUND);
    // 没有密码默认给随机一个
    const password = get(data,'password', randomStr())
    user.password = password;
    await user.save();
    return password;
  }
}
